# -* perl *-

## progressive multiple alignment

use MLocarna;
use strict;

##------------------------------------------------------------
## options

# we use the GetOpt module in a mode that allows
# to pass a list of options quite easilz to locarna
# while handling other options explicitely

use Getopt::Long;
use Pod::Usage;

my $help;
my $man;

my $quiet;
my $verbose;

my $noLP=1;
my $no_lonely_pair;
my $lonely_pair;

my $tree_file;
my $tgtdir;

my $mfasta_file;

my $skip_pp=0; ## skip the computation of pair probabilities for files
               ## that exist already 

my $ext_library;

my $iterate;
my $iterations;

my $min_prob=0.0005;
my $tree_min_prob=0.01;

my $stacking;

my $moreverbose;

my $exp_prob;

my $ignore_constraints;

my $consistency_transformation;



Getopt::Long::Configure("no_ignore_case");

my %options_hash=
    (
     "tree-min-prob" => \$tree_min_prob,
     "noLP" => \$no_lonely_pair,
     "LP" => \$lonely_pair,
     
     "min-prob" => \$min_prob,
     "tree-min-prob" => \$tree_min_prob,

     "stacking" => \$stacking,

     "extlib" => \$ext_library,
     
     "skip-pp" => \$skip_pp,
     
     "iterate" => \$iterate,
     "iterations" => \$iterations,

     "ignore-constraints" => \$ignore_constraints,

     "consistency-transformation" => \$consistency_transformation,
     
     "treefile" => \$tree_file,
     "outdir" => \$tgtdir,
     
     "verbose" => \$verbose,
     "moreverbose" => \$moreverbose,
     "quiet" => \$quiet,
     
     "help"=> \$help,
     "man" => \$man
    );


my @options = #options that are not or not automatically passed to locarna
    (
     "noLP",
     "LP",
     
     "min-prob|p=f",
     "tree-min-prob|P=f",
     
     "stacking",
     
     "extlib",
     
     "skip-pp",
     
     "iterate",
     "iterations=i",
     
     "ignore-constraints",
     "consistency-transformation",

     "treefile=s",
     "outdir|tgtdir=s",
     
     "verbose",
     "moreverbose",
     "quiet",
     
     "help",
     "man"
    );

my @autopass_arg_options = #options with argument that are passed to locarna
    (
     "match|m=i",
     "mismatch|M=i",
     "indel|i=i",
     "indel-open=i",
     "ribosum-file=s",
     "struct-weight|s=i",
     "tau|t=i",
     "exclusion|E=i",
     "exp-prob|e=f",
     
     "struct-local=s",
     "sequ-local=s",
     "width=i",
     "max-diff-am|D=i",

     "probcons-file=s",
     "match-prob-method=i",
     "gap-opening=i",
     "gap-extension=i",
     "temperature=i",
     
     "mea-alpha=i",
     "mea-beta=i",
     "mea-gamma=i"
    );

my @autopass_noarg_options = ( "mea-alignment" );

# -----------------------------------
# some default values
$options_hash{"struct-weight"}=120;
$options_hash{"max-diff-am"}=30;


GetOptions( \%options_hash, @options, @autopass_arg_options, @autopass_noarg_options )
    || pod2usage(2);


pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

if ($moreverbose) {
    $verbose=1;
}

if (defined($lonely_pair) && defined($no_lonely_pair)) {
    print STDERR "Only one of the options --noLP and --LP can be defined at a time.\n";
    pod2usage(1);
}
if (defined($lonely_pair)) {
    $noLP=0;
}
if (defined($no_lonely_pair)) {
    $noLP=1;
}

if (exists $options_hash{"exp-prob"}) {
    $exp_prob=$options_hash{"exp-prob"};
}

## construct parameter string for locarna

my $locarna_params="";
my $locarna_params_tree="";

$locarna_params.=" -v" if ($verbose);

$locarna_params.=" --noLP" if ($noLP);

# $locarna_params.=" --ignore-constraints" if ($ignore_constraints);

foreach my $opt (@autopass_arg_options) {
    $opt =~ s/[|=].*//;
    
    if (exists $options_hash{$opt}) {
	$locarna_params .= " --$opt $options_hash{$opt}";
    }
}

foreach my $opt (@autopass_noarg_options) {
    $opt =~ s/[|].*//;
    
    if (exists $options_hash{$opt}) {
	$locarna_params .= " --$opt";
    }
}


$locarna_params_tree=$locarna_params;

## special treatment of min-prob and tree-min-prob

$locarna_params .= " --min-prob $min_prob" if (defined($min_prob));
if (defined($tree_min_prob)) {
    $locarna_params_tree.=" --min-prob $tree_min_prob";
} else {
    $locarna_params_tree.=" --min-prob $min_prob" if (defined($min_prob));
}

$locarna_params .= " --stacking" if ($stacking);

## ----------------------------------------
# mfasta file

if ($#ARGV == 0) {
    $mfasta_file = $ARGV[0];
}else {
    pod2usage(1);
}


## target/output directory
if (!defined($tgtdir)) {
    $tgtdir = $mfasta_file;
    $tgtdir =~ s/\.mfasta$|\.fasta$|\.fa$//;
    $tgtdir .= ".out";
}


if ($verbose) {
    print "Locarna Parameter for progressive alignment: $locarna_params\n";
    print "Locarna Parameter for guide tree construction: $locarna_params_tree\n";
}



##------------------------------------------------------------


my $intermediate_name_base="intermediate";
my $intermediate_name_counter=0;

sub new_intermediate_name {
    $intermediate_name_counter++;
    return "$intermediate_name_base$intermediate_name_counter";
}

my $node_sym="\$\$nodesym";


##------------------------------------------------------------

sub parse_tree;
sub parse_dp_ps;
sub subtract_list;


### ============================================================================### ============================================================================
### ============================================================================
## ---------------------------------------------------------------------------
# main program
#

if (! -e $tgtdir) {
    mkdir $tgtdir || die "Cannot create $tgtdir";
} else {
    print STDERR "Warning: $tgtdir exists already.\n" if ($verbose);
}

## ----------------------------------------
# read the mfasta input file
my %mfasta = parse_mfasta_constraints($mfasta_file);

my @names = keys %mfasta;     # get all keys in the hash
@names = grep {!/\#/} @names; # keep only the sequence names (without constraint strings)


my %filenames;


chdir $tgtdir;

my $tmpname="_tmp_"; # valid file name, at max 12 characters
my $tmpprefix="$tmpname"; # could be a absolute path, but there's no need to

## ----------------------------------------
# compute probability dot-plots
# (if wanted or if necessary)
#
print "Compute pair probs ...\n" unless $quiet;

for my $name (@names) {
    my $RNAfold_args="";
    $RNAfold_args=">/dev/null" unless $verbose;
    
    if ( $skip_pp && -e $name ) {
	# skip
    } else {
	## ----------------------------------------
	# partition folding
	#
	
	if ((!$ignore_constraints) && exists $mfasta{"$name\#S"}) {
	    # if a structure constraint is defined do constraint folding
	    my $constraint_string = $mfasta{"$name\#S"};
	    system "printf '>$tmpname\n$mfasta{$name}\n$constraint_string' | RNAfold -p2 $RNAfold_args -C";
	} else {
	    # else do normal pf
	    system "printf '>$tmpname\n$mfasta{$name}' | RNAfold -p2 $RNAfold_args";
	}
	
	my $constraints;
	if (! $ignore_constraints) {
	    $constraints = $mfasta{"$name\#C"};
	}
	
	convert_dp_to_pp_with_constraints("$tmpname\_dp2.ps",$name,$name,$mfasta{$name},$constraints);
	# convert was before a simple rename:
        #rename "$shortname\_dp.ps", "$name"
	#    || die "Cannot rename $shortname\_dp.ps to $name\n";
	
	unlink "$tmpname\_ss.ps" || die "Cannot delete $tmpname\_ss.ps";
 	unlink "$tmpname\_dp.ps" || die "Cannot delete $tmpname\_dp.ps";
 	unlink "$tmpname\_dp2.ps" || die "Cannot delete $tmpname\_dp2.ps";
    }
}

my %seq_lengths;
foreach my $name (@names) {
    my %aln = parse_pp_file_aln($name);
    $seq_lengths{$name} = aln_length(\%aln);
}


## ----------------------------------------
## generate base match probabilities for all sequence pairs
#  (if required)
#
my %bmprobs; # base match probabilities

if ($consistency_transformation) {

    print "Compute base match probabilities (for consistency based transformation).\n" unless $quiet;
    
    %bmprobs = compute_basematch_probs(\%seq_lengths,\@names);
    
    # print_k_dim_hash(\%bmprobs,4,"");

    print "Consistency transform base match probabilities.\n" unless $quiet;

    %bmprobs = consistency_transform(\%bmprobs);

    #print_k_dim_hash(\%bmprobs,4,"");

}



## ----------------------------------------
# generate guide tree (if not given)
#

my $tree;

if (defined($tree_file)) {
    open(TREE,"$tree_file") || die "Cannot read $tree_file for reading tree\n";
    $tree=<TREE>;
    chomp($tree);
    close TREE;
} elsif ($#names == 1) { # for only two sequences, tree is unique
    $tree="($names[0],$names[1])";
} else {
    # --------------------------------------------------
    # compute score matrix for guide tree
    #
    
    my @score_matrix;
    my @pairwise_alns;
    
    ## compute pairwise scores in score_matrix
    
    print "Compute pairwise alignments ... \n" unless $quiet;

    my $num=0;
    for (my $a=0; $a<=$#names; $a++) {
	for (my $b=0; $b<$a; $b++) {
	    $num++;
	    
	    printf 
		"Align %d/%d : $names[$a] and %d/%d: $names[$b] (%d/%d)\n",
		$a+1,$#names+1,$b+1,$a,$num,($#names*($#names+1))/2 if $verbose;

	    my @aln;
	    if ($consistency_transformation) {
		@aln =
		    compute_alignment_from_dps_cbt($names[$a],
						   $names[$b],
						   $locarna_params_tree,
						   $bmprobs{$names[$a]}{$names[$b]}
		    );
	    } else {
		@aln = compute_alignment_from_dps($names[$a],
						  $names[$b],
						  $locarna_params_tree,
						  $tmpprefix
						  ); 
	    }
	    
	    $pairwise_alns[$a][$b] = [ @aln ];
	}
    }
    
    for (my $a=0; $a<=$#names; $a++) {
	for (my $b=0; $b<$a; $b++) {
	    
	    my @aln = @{ $pairwise_alns[$a][$b] };
	    
	    # print "Alig. of $names[$a] --- $names[$b]:\n@aln\n";
	    
	    $aln[0] =~ /Score: ([\d\.\-]+)/ || die "Cannot extract score from $aln[0].\n";
	    my $score=$1;
	    
	    $score_matrix[$a][$b] = $score;
	    $score_matrix[$b][$a] = $score;
	}
    }

    open(MAT,">result.matrix") || die "Cannot write score matrix\n";
    for (my $a=0; $a<=$#names; $a++) {
	for (my $b=0; $b<=$#names; $b++) {
	    if ($a==$b) {print MAT "     0 ";} else { 
		printf MAT "%6d ",$score_matrix[$a][$b];
	    }
	}
	print MAT "\n";
    }
    close MAT;
    # system("cat $tgtdir/matrix");
    
    
    if ($ext_library) {
	extend_library(\@pairwise_alns);
    }


    $tree = compute_tree(\@names,\%mfasta,\@score_matrix);    

} # end generation of tree

open(TREE,">result.tree")|| die "Cannot write to $tgtdir/result.tree";
print TREE "$tree\n";
close TREE;

my @tree_postorder = parse_tree($tree);

# print "TREE: <@tree_postorder>\n";
    

# print "@tree_postorder\n";

## if we read the tree from a file, then we have to shorten the names,
## and substitute special characters
if (defined($tree_file)) {
    for (my $i=0; $i<=$#tree_postorder; $i++) {
	if ( $tree_postorder[$i] ne $node_sym ) {
	    $tree_postorder[$i] =  substr $tree_postorder[$i],0,16;
	    $tree_postorder[$i] =~ s/[^a-zA-Z\d]/_/g;
	}
    }
}


## ----------------------------------------
# progressive alignment along the guide tree
#
my @stack=();

for my $item (@tree_postorder) {   
    if ($item eq $node_sym) {
	# align two topmost stack entrys and push result
	my $op1 = $stack[$#stack-1];
	my $op2 = $stack[$#stack];
	
	$#stack -= 2;
	
	my $intermediate_name = new_intermediate_name();

	if (! $consistency_transformation) {
	    call_locarna($op1,$op2,$intermediate_name);
	} else {
	    call_locarna_cbt($op1,$op2,$intermediate_name);
	}
	
	push @stack, "$intermediate_name.pp";
	
    } else {
	push @stack, $item;
    }
}

if ($#stack!=0) {
    print STDERR "Strange error occured! Exit.\n";
    print "STACK: @stack\n";
    exit -1;
}

## ----------------------------------------
# copy results to files result_prog.aln and result_prog.pp
#
(my $current_alignment_name = $stack[-1] ) =~ s/.pp$//;
system "cp", "$current_alignment_name.aln", "result_prog.aln";
system "cp", "$current_alignment_name.pp", "result_prog.pp";


### ============================================================
### iterative refinement

##
## 1.) from the guide tree get all bi-partitionings of seq-names,
##     that are separated by tree edges 
## 2.) start with the final multiple alignment from the progressive phase
##     as the current ma
## 3.) for each partitioning
##     a) extract the multiple alignments of both partitions out of the current ma
##     b) generate new pp files with newly computed consensus pairprobs
##     c) align both paritions in order to form a new, better ma
##
##
## Details:
## ad 3a) read the current ma and generate the two sub-alignments
##        by projection, delete only-gap columns
## ad 3b) read all pp files of the single sequences in the sub-alignment
##        and compute consensus dot-plot by averaging probs
##        (geometric or arithmetic mean).
##        This requires two compute the projection of the single sequences
##        to their lines in the alignment.
##
## Handling of constraints?
## merge also constraint lines from the single pp files
##

##
## UNCLEAN: computation of consensus dot-plots, reading and writing pp-files, ...
## is duplicated in C++-code and perl-code.
## ==> The multiple alignment computation should be migrated
## to C++-code, i.e. mlocarna should be rewritten in C++.
##

if (defined($iterate) && ! defined($iterations)) {
    $iterations=1;
}

for (my $i=0; $i<$iterations; $i++) {
    print "\n" unless $quiet;
    print "============================\n" unless $quiet;
    print "--- Iterative Refinement ---\n" unless $quiet;
    print "============================\n" unless $quiet;
    print "\n" unless $quiet;

    my @partitions = tree_partitions(@tree_postorder);
    
    foreach my $namesA_ref (@partitions) {
	my @namesA = @{ $namesA_ref };
	my @namesB = subtract_list( \@names, \@namesA );
		
	print "Realign split @namesA --- @namesB\n" unless $quiet;
	
	my %aln = parse_aln("$current_alignment_name.aln");
	
	my $intermediate_name = new_intermediate_name();
	
	my $intermediate_nameA = $intermediate_name."A.pp";
	my $intermediate_nameB = $intermediate_name."B.pp";
	
	write_pp_projected(\%aln,\@namesA,$intermediate_nameA);
	write_pp_projected(\%aln,\@namesB,$intermediate_nameB);
	
	call_locarna($intermediate_nameA,$intermediate_nameB,$intermediate_name);
	$current_alignment_name = $intermediate_name;
    }
}


## ----------------------------------------
# copy results to files result.aln and result.pp
#
system "cp", "$current_alignment_name.aln", "result.aln";
system "cp", "$current_alignment_name.pp", "result.pp";


### ============================================================================
### ============================================================================
### ============================================================================
### SUBS
###

## compute basematch probs for all pairs of names
## we assume that for each name there is a pp file
## for computing the probabilities, locarna is called with
## option --write-basematch-probs
## returns a hash of "sparse matrices"
sub compute_basematch_probs {
    my ($seq_lengths_ref,$names_ref) = @_;

    my %seq_lengths = %{ $seq_lengths_ref };
    my @names = @{ $names_ref };
    
    my %bmprobs;
    
    for my $nameA (@names) {
	for my $nameB (@names) {
	    if ($nameA ne $nameB) {
		if (exists $bmprobs{$nameB}{$nameA}) {
		    # basematch probs are symmetric. So if we have it already then use it.

		    my %subhash = transpose_sparsematrix( $bmprobs{$nameB}{$nameA} );
		    $bmprobs{$nameA}{$nameB} = \%subhash;
		    
		} else {
		    # compute the base match probabilities for the pair
		    my $tmpfile="/tmp/mlocarna.$$.bmps";
		    
		    my $cmd = "locarna $nameA $nameB $locarna_params --write-match-probs $tmpfile";
		    $cmd .= " >/dev/null" unless $moreverbose;
		    print "$cmd\n" if $moreverbose;
		    system($cmd)==0 || die "Cannot call locarna and write to $tmpfile";
		    
		    my %mat = read_sparse_matrix($tmpfile);
		    
		    $bmprobs{$nameA}{$nameB} = { %mat };
		    
		    unlink "$tmpfile" || die "Cannot remove $tmpfile";
		}
	    }
	}
    }

    for my $name (@names) {
	my $len = $seq_lengths{$name};

	for (my $i=1; $i<=$len; $i++) {
	    $bmprobs{$name}{$name}{$i}{$i}=1.0;
	}
    }
    
    return %bmprobs;
}

## get a hash indexed by name pairs of "sparse matrices" = match probs for all sequence pairs
## returns the transformed hash
sub consistency_transform {
    my ($h_ref) = @_;

    my %h = %{ $h_ref };

    # print_k_dim_hash(\%h,4,"");
    
    my %th;
    
    my @names = keys %h;
    
    foreach my $nameA (@names) {
	foreach my $nameB (@names) {
	    if ($nameA eq $nameB) {
		$th{$nameA}{$nameB} = $h{$nameA}{$nameB};
	    } else {
		if (exists $th{$nameB}{$nameA}) {
		    my %subhash = transpose_sparsematrix( $th{$nameB}{$nameA} );
		    $th{$nameA}{$nameB} = \%subhash;
		} else {

		    foreach my $nameC (@names) {
			
			# transform via $nameC and add to matrix AB
						
			my %matrixAB = consistency_transform_single( $h{$nameC}{$nameA} , $h{$nameC}{$nameB} );

			if (exists $th{$nameA}{$nameB}) {
			    my %sum = add_sparsematrix ( $th{$nameA}{$nameB}, \%matrixAB );
			    
			    $th{$nameA}{$nameB} = \%sum;
			} else {
			    $th{$nameA}{$nameB} = \%matrixAB;
			}
		    }

		    my %m = divide_sparsematrix( $th{$nameA}{$nameB}, $#names+1 );
		    $th{$nameA}{$nameB} = \%m;
		    
		}
	    }
	}
    }
    
    return %th;
}

sub print_k_dim_hash($$$) {
    my ($h_ref,$k,$s) = @_;
    my %h = %{ $h_ref };
    
    foreach my $i (sort keys %h) {
	print "$s$i:";
	if ($k==1) {
	    print "$h{$i}\n";
	} else {
	    print "\n";
	    print_k_dim_hash( $h{$i},$k-1,"$s\t" );
	}
    }
}


sub consistency_transform_single($$) {
    my ($mCA_ref,$mCB_ref) = @_;

    my %mAB;
    
    my %mCA = %{ $mCA_ref };
    my %mCB = %{ $mCB_ref };
    
    # ATTENTION: actually we need to traverse the cut set of lists keys mCA and keys mCB
    foreach my $k (keys %mCA) {
	if (exists $mCA{$k} && exists $mCB{$k}) {
	    my %hA = %{ $mCA{$k} };
	    my %hB = %{ $mCB{$k} };
	    
	    foreach my $i (keys %hA) {
		foreach my $j (keys %hB) {
		    $mAB{$i}{$j} += $hA{$i} * $hB{$j}
		}
	    }
	}
    }
    return %mAB;
}

sub add_sparsematrix($$) {
    my ($m1_ref,$m2_ref) = @_;

    my %m1 = %{ $m1_ref };
    my %m2 = %{ $m2_ref };

    my %m3 = %m1;

    foreach my $i (keys %m2) {
	foreach my $j (keys %{ $m2{$i} }) {
	    $m3{$i}{$j} += $m2{$i}{$j}; 
	}
    }
    
    return %m3;
}

sub divide_sparsematrix($$) {
    my ($m_ref,$divisor) = @_;

    my %m = %{ $m_ref };
    my %r;

    foreach my $i (keys %m) {
	foreach my $j (keys %{ $m{$i} }) {
	    $r{$i}{$j} = $m{$i}{$j} / $divisor; 
	}
    }

    return %r;    
}


# read from a sparse matrix file as written by locarna --write-match-probs 
sub read_sparse_matrix($) {
    my ($file) = @_;

    local *SM_IN;
    
    open(SM_IN,$file);

    my %h;
    
    while( my $line=<SM_IN> ) {
	if ( $line =~ /(\d+) (\d+) ([\d.e-]+)/ ) {
	    $h{$1}{$2} = $3;
	}
    }
    
    close SM_IN;

    return %h;
}

sub write_sparse_matrix($$) {
    my ($m_ref,$file)=@_;
    my %m = %{ $m_ref };

    local *SM_OUT;
    
    open(SM_OUT,">$file") || die "Cannot write to $file."; 

    foreach my $i (keys %m) {
	foreach my $j (keys %{ $m{$i} }) {
	    print SM_OUT "$i $j $m{$i}{$j}\n"; 
	}
    }
    
    close SM_OUT;
}


# transpose a sparse matrix implemented as hash of hashs
sub transpose_sparsematrix($) {
    my ($m_ref) = @_;

    my %m = %{ $m_ref };
    
    my %r;
    
    foreach my $i (keys %m) {
	foreach my $j (keys %{ $m{$i} }) {
	    $r{$j}{$i} = $m{$i}{$j}; 
	}
    }
    
    return %r;
}


## calls locarna on inA and inB and writes to $tgt.pp and $tgt.aln
sub call_locarna($$$) {
    my ($inA,$inB,$tgt) = @_;
    
    print "--------------------\nAlign $inA with $inB --> $tgt\n" unless $quiet;
    my $cmd="locarna $inA $inB $locarna_params --pp=$tgt.pp --clustal=$tgt.aln";
    if (! $moreverbose) { $cmd.=">/dev/null"; }
    
    print "$cmd\n" if ($verbose);
    print "\n"  unless $quiet;
    system($cmd)==0
	|| die "Cannot execute locarna\n";
    
    system "cat $tgt.aln" unless ($quiet);
}

sub average_basematch_probs($$) {
    my ($alnA_ref,$alnB_ref) = @_;

    my %alnA = %{ $alnA_ref };
    my %alnB = %{ $alnB_ref };
    
    my @namesA = grep !/#/, keys %alnA;
    my @namesB = grep !/#/, keys %alnB;

    my %r; # result matrix

    for my $nameA (@namesA) {
	for my $nameB (@namesB) {
	    # projecterei

	    my @posmapA = project_seq($alnA{$nameA});
	    my @posmapB = project_seq($alnB{$nameB});
	    
	    my %m = %{ $bmprobs{$nameA}{$nameB} };

	    # averagerei
		
	    foreach my $i (keys %m) {
		foreach my $j (keys %{ $m{$i} }) {
		    $r{$posmapA[$i]}{$posmapB[$j]} += $m{$i}{$j};
		}
	    }
	}
    }

    my $nm = ($#namesA+1)*($#namesB+1);
    
    return divide_sparsematrix(\%r,$nm);
}

sub call_locarna_cbt($$$) {
    my ($inA,$inB,$tgt) = @_;

    # get alns

    my %alnA = parse_pp_file_aln($inA);
    my %alnB = parse_pp_file_aln($inB);
    
    my %m = average_basematch_probs( \%alnA, \%alnB );
    
    my $tmpfile="/tmp/mlocarna.$$.bmps";

    write_sparse_matrix(\%m,$tmpfile);
    
    print "--------------------\nAlign $inA with $inB --> $tgt\n" unless $quiet;
    my $cmd="locarna $inA $inB $locarna_params --pp=$tgt.pp --clustal=$tgt.aln --read-match-probs=$tmpfile";
    if (! $moreverbose) { $cmd.=" >/dev/null"; }
    
    print "$cmd\n" if ($verbose);
    print "\n";
    system($cmd)==0
	|| die "Cannot execute locarna\n";
    
    system "cat $tgt.aln" unless ($quiet);
}


# read multiple alignment in aln format
sub parse_aln($) {
    my ($aln_filename) = @_;
    local *ALN_IN;
    
    open(ALN_IN,$aln_filename) || die "Can not read aln file $aln_filename\n";
    
    my %aln;
    
    my $line;

    if (($line=<ALN_IN>) !~ /CLUSTAL/) {
	print STDERR "$aln_filename not in clustal aln-format.\n";
	exit(-1);
    }
    
    while ($line = <ALN_IN>) {
	if ($line =~ /^([^\s]+)\s+(.+)/) {
	    $aln{$1} .= $2;
	}
    }
    
    close ALN_IN;
        
    return %aln;
}

# project multiple alignment to subset of names
# cleans only-gap columns
sub project_aln {
    my ($aln_ref,$names_ref) = @_;

    my %aln = %{ $aln_ref };
    my @names = @{ $names_ref };

    my %alnP;

    foreach my $name (@names) {
	$alnP{$name} = $aln{$name};
    }
    
    my $len = aln_length(\%alnP);
    
    for (my $i=0; $i<$len; ) {
	my $allgap=1;
	foreach my $name (keys %alnP) {
	    my $row = $alnP{$name};
	    # print "$name $row $i\n";
	    if (! is_gap(substr($row,$i,1))) {
		$allgap=0;
		last;
	    }
	}
	if ($allgap) {
	    # remove alignment column i
	    foreach my $name (keys %alnP) {
		substr($alnP{$name},$i,1) = "";
	    }
	    $len--;
	} else {
	    $i++;
	}
    }

    
    return %alnP;
}

sub aln_length($) {
    my ($aln_ref) = @_;
    my %aln = %{ $aln_ref };
    my @ks = keys %aln;
    # print "$ks[0], len $aln{$ks[0]} " . length( $aln{$ks[0]} ). "\n";
    return length( $aln{$ks[0]} );
}

sub aln_size($) {
    my ($aln_ref) = @_;
    my %aln = %{ $aln_ref };
    my @ks = keys %aln;
    @ks = grep !/#/,@ks;
    return $#ks+1;
}

# write pp file from aln and consensus dp, write only probs >= min_prob
sub write_pp($$$$) {
    my ($file,$aln_ref, $cons_ref, $min_prob) = @_;
    my %aln = %{ $aln_ref };
    my @cons = @{ $cons_ref };
    
    local *OUT;
    open(OUT, ">$file") || die "Cannot open $file for writing\n";

    print OUT "SCORE: 0\n\n";
    
    foreach my $name (keys %aln) {
	printf OUT "%-20s %s\n", $name, $aln{$name};
    }
    
    
    print OUT "\n#\n";

    my $len = aln_length(\%aln);
    
    for (my $i=1; $i<=$len; $i++) {
	for (my $j=$i+1; $j<=$len; $j++) {
	    if (defined $cons[$i][$j]) {
		my $p = $cons[$i][$j];
		if ($p >= $min_prob) {
		    print OUT "$i $j $p\n";
		}
	    }
	}
    }
    
    close OUT;
}



# write pp file out of the alignment projected to
# the given names
# generates consensus dotplot and constraints
# out of single pp files
sub write_pp_projected($$$) {
    my ($aln_ref, $names_ref, $tgt_name) = @_;
    my %aln = %{ $aln_ref };
    my @names = @{ $names_ref };
    
    my %alnP = project_aln(\%aln,\@names);

    ## compute the consensus dot plot
    my @consensus_dp; # 2D array
    my @size; # number of added values for pos $i,$j
    
    foreach my $name (@names) {
	my @posmap = project_seq($alnP{$name});
	my %name_aln = parse_pp_file_aln($name);	
	my %name_pairprobs = parse_pp_file_pairprobs($name);	
	
	if ( exists $name_aln{"#C"} && (!$ignore_constraints)) {
	    my $constraints = $name_aln{"#C"};
	    print "WARNING: constraints ignored in iterative alignment.\n";
	}

	# for computing the geometric mean, we don't allow 0 values,
	# since this extinguishs the probability completely!
	# therfore we use some very small 'background' probability
	# if $p is too low.
	#
	# In principle, copy the magic spell from alignment.cc
	#
	my $minimal_p = $min_prob*0.75;
	if (defined($exp_prob) && $exp_prob<$minimal_p ) {
	    $minimal_p = $exp_prob;
	}
		
	for my $pair (keys %name_pairprobs) {
	    $pair =~ /(\d+) (\d+)/;
	    my $i=$1;
	    my $j=$2;
	    my $p = $name_pairprobs{$pair};

	    if ($p<$minimal_p) {$p=$minimal_p;}
	    
	    $consensus_dp[$posmap[$i]][$posmap[$j]] += log($p);
	    $size[$posmap[$i]][$posmap[$j]] ++;
	}
    }
    

    # normalize consensus_dp
    my $len  = aln_length(\%alnP);
    
    for (my $i=1; $i<=$len; $i++) {
	for (my $j=$i+1; $j<=$len; $j++) {
	    if (defined($consensus_dp[$i][$j])) {
		$consensus_dp[$i][$j] = exp($consensus_dp[$i][$j] / $size[$i][$j]);
	    }
	}
    }
    
    write_pp($tgt_name,\%alnP,\@consensus_dp,$min_prob);
    
    return;
}

# produce an array that yields for each sequence position the corresponding
# position in the given alignment string
sub project_seq($) {
    my ($alig_str) = @_;
    
    my @posmap;
    
    my $len=length($alig_str);
    
    my $j=1;
    for (my $i=1; $i<=$len; $i++) {
	if ( ! is_gap(substr($alig_str,$i-1,1)) ) {
	    $posmap[$j]=$i;
	    $j++;
	}
    }
    
    return @posmap;
}

sub is_gap($) {
    my ($s)=@_;
    return ($s eq "-") || ($s eq "~"); 
}


sub subtract_list {
    my ($l1ref, $l2ref) = @_;

    my @l1 = @{ $l1ref };
    my @l2 = @{ $l2ref };

    my @res;
    
    foreach my $x (@l1) {
	my $found=0;
	foreach my $y (@l2) {
	    if ($x eq $y) {
		$found=1;
		last;
	    }
	}
	if ($found==0) { push @res, $x; }
    }
    return @res;
}

## generate the partitions out of the postorder tree
sub tree_partitions(@) {
    my @tree_postorder = @_;

    my @result;
    
    my $stack;
    
    for my $item (@tree_postorder) {
	if ($item eq $node_sym) {
	    my @op1 = @{ $stack[-2] };
	    my @op2 = @{ $stack[-1] };
	    
	    $#stack-=2;
	    
	    #push @result, \@op1;
	    #push @result, \@op2;
	    
	    my @op12 = (@op1, @op2);
	    
	    push @stack, \@op12;
	} else {
	    push @stack, [ $item ];
	}
	push @result, $stack[-1];
    }
    $#result-=2; # the last is empty, the one before symmetric
    return @result;
}

sub parse_tree {
    my ($tree) = @_;

    my @list;
    
    $tree =~ s/:[\d.e-]+//g;
    
    print "Tree: $tree\n" unless $quiet;
    
    my $brcount=0;
    for (my $i=0; $i< length $tree; $i++) {
	my $c=substr $tree,$i,1;
	
	if ($c eq "(") {
	    $brcount++;
	} elsif ($c eq ")") {
	    $brcount--;
	    if ($brcount<0) {
		print STDERR "Parse error in tree.";
		exit -1;
	    }
	    
	    push @list, $node_sym;
	} elsif ($c eq ",") {
	    ## ignore, although we could do syntax checking
	} else {
	    my $rest=substr $tree,$i;
	    $rest =~ /^([^(),]+)/;
	    my $token = $1;
	    $i+=(length $token)-1;
	    push @list, $token;
	}
    }

    return @list;
}

## compute a WPGMA-tree from the given mfasta-hash
sub compute_tree ($$$) {
    # print "Compute WPGMA-Tree ...\n";

    my ($names_ref, $mfasta_ref, $score_matrix_ref) = @_;
    
    my @names = @$names_ref;
    my %mfasta = %$mfasta_ref;
    my @score_matrix = @$score_matrix_ref;
    
    
    ## compute tree by wpgma applied to score matrix
    
    my @clusters; # a list of the clusters
    my @trees; # a list of sub-trees
    my @cluster_sizes;

    for (my $i=0; $i<=$#names; $i++) {
	$clusters[$i]=$i;
	$trees[$i]=$names[$i];
	$cluster_sizes[$i]=1;
    }
    
    
    my $NEG_INFINITY=-1e10;
    
    while ($#clusters>0) {
	
	#print "@clusters\n";
	#print "@trees\n";
	#print "@cluster_sizes\n";
	
        ## find the nearest two clusters
	
	my $max_i;
	my $max_j;
	my $max_score=$NEG_INFINITY;
	for (my $i=0; $i<=$#clusters; $i++) {
	    for (my $j=$i+1; $j<=$#clusters; $j++) {
		my $score=$score_matrix[$clusters[$i]][$clusters[$j]];
		if ($score > $max_score) {
		    $max_i=$i; 
		    $max_j=$j;
		    $max_score=$score;
		}
	    }
	}
	
	print "Merge $clusters[$max_i] $clusters[$max_j] --> $clusters[$max_i]\n" if $verbose;
	
	## recompute distances
	my $cluster_i = $clusters[$max_i];
	my $cluster_j = $clusters[$max_j];

	$clusters[$max_j] = $clusters[$#clusters];
	$clusters[$max_i] = $clusters[0];
	$clusters[0] = $cluster_i;
	
	for (my $i=1; $i<$#clusters; $i++) {
	    $score_matrix[$clusters[0]][$clusters[$i]]
		= ($cluster_sizes[$cluster_i] * $score_matrix[$cluster_i][$clusters[$i]] + $cluster_sizes[$cluster_j] * $score_matrix[$cluster_j][$clusters[$i]]) / ($cluster_sizes[$cluster_i]+$cluster_sizes[$cluster_j]);
	    $score_matrix[$clusters[$i]][$clusters[0]] = 
		$score_matrix[$clusters[0]][$clusters[$i]];
	}
	
	my $new_tree="($trees[$cluster_i],$trees[$cluster_j])";
	print "Merge $trees[$cluster_i] $trees[$cluster_j] --> $new_tree\n" if $verbose;
	
	$trees[$clusters[0]]=$new_tree;
	
	#$trees[$clusters[$#clusters]]="undef";
	#$cluster_sizes[$clusters[$#clusters]]=0;
    
	$cluster_sizes[$clusters[0]]=$cluster_sizes[$cluster_i]+$cluster_sizes[$cluster_j]; 
	
	$#clusters--;
    }


    print "Computed-tree: $trees[$clusters[0]]\n" if $verbose;

    return $trees[$clusters[0]];
}



sub seqpos_to_alipos {
    my ($ali) = @_;
    
    my @res;

    my $seqpos=0;
    my $alipos=0;
    
    for my $c (split //, $ali) {
	if ($c =~ /[ACGUT]/) {
	    push @res,$alipos;
	    $seqpos++;
	}
	$alipos++;
    }

    return @res;
}

sub alipos_to_seqpos {
    my ($ali) = @_;
    
    my @res;

    my $seqpos=0;
    my $alipos=0;
    
    for my $c (split //, $ali) {
	if ($c =~ /[ACGUT]/) {
	    push @res,$seqpos;
	    $seqpos++;
	} else {
	    push @res,-1;
	}
	$alipos++;
    }

    return @res;
}


sub parse_dp_ps {
    my ($filename) = @_;
    local *IN;
    
    open(IN,$filename) || die "Cannot read $filename for parsing as dp-ps file.\n";
    
    my $seq="";
    my @pairprobs;
    
    while (my $line=<IN>) {
	if ($line =~ /^\/sequence \{ \(/) {
	    while (defined($line = <IN>) && ($line !~  /\} def/  ))  {
		chomp $line;
		$line =~ s/\\$//;
		$seq .= $line;
	    }
	    print "parse_dp_ps $filename: $seq\n";
	}
	
	if ($line =~ /(\d+)\s+(\d+)\s+(\S+)\s+ubox/) {
	    $pairprobs[$1-1][$2-1]=$3*$3;
	}
    }
    
    close IN;
    
    $seq ne "" || die "Empty sequence in dp.ps file $filename\n";
    
    return ($seq,\@pairprobs);
}

sub extend_library {
    my ($pairwise_alns_ref) = @_;
    my @pairwise_alns=@{ $pairwise_alns_ref };
    
    
    ### ------------------------------------------------------------
    ## compute "extended pairprobability library"
    #
    
    my @pairprobs;
    my @sequences;
    
    for (my $a=0; $a<=$#names; $a++) {
	my $seqA;
	my $pairprobsA_ref;
	($seqA,$pairprobsA_ref) = parse_dp_ps("$names[$a]");
	
	$sequences[$a] = $seqA;
	$pairprobs[$a] = [ @{ $pairprobsA_ref } ];
    }
    
    print "Compute extended pairprobability library ...\n";
    
    ## make a deep copy of pair probabilities
    #
    my @ext_pairprobs;
    for (my $a=0; $a<=$#names; $a++) {
	my @eppA;
	my @ppA = @{ $pairprobs[$a] };
	for my $i ( 0..$#ppA ) {
	    if (defined $ppA[$i]) {
		my @ppAa = @{ $ppA[$i] };
		for my $j ( 0..$#ppAa ) {
		    if (defined $ppAa[$j]) {
			$eppA[$i][$j] = $ppA[$i][$j];
		    }
		}
	    }
	}
	$ext_pairprobs[$a] = [ @eppA ];
    }
    
    for (my $a=0; $a<=$#names; $a++) {
	print "Compute ext lib for $names[$a]\n";
	
	my @ext_pairprobsA = @{ $ext_pairprobs[$a] }; # we accummulate the extension!
	
	for (my $b=0; $b<=$#names; $b++) {
	    if ($a != $b) {
		
		my $aliA; # alignment-string for sequence a
		my $aliB; # alignment-string for sequence b
		
		if ($a>$b) {
		    my @alnAB = @{ $pairwise_alns[$a][$b] };
		    ($aliA,$aliB) = parse_alignment($names[$a],$names[$b],\@alnAB);
		} else {
		    my @alnBA = @{ $pairwise_alns[$b][$a] };
		    ($aliB,$aliA) = parse_alignment($names[$b],$names[$a],\@alnBA);
		}
		
		my @sp2apA = seqpos_to_alipos($aliA);
		my @sp2apB = seqpos_to_alipos($aliB);
		
		my @ap2spA = alipos_to_seqpos($aliA);
		my @ap2spB = alipos_to_seqpos($aliB);
		
		my $lengthA=$#sp2apA+1;
		my $lengthB=$#sp2apB+1;
		
		print "  $a:$names[$a] ($lengthA) --- $b:$names[$b]\n";
		
		my @pairprobsB = @{ $pairprobs[$b] };
		
		for (my $iA=0; $iA<$lengthA; $iA++) {
		    for (my $jA=$iA+1; $jA<$lengthA; $jA++) {
			my $pA = (defined $ext_pairprobsA[$iA][$jA])?$ext_pairprobsA[$iA][$jA]:$min_prob;
			if ($pA > $min_prob) {
			    my $iB = $ap2spB[$sp2apA[$iA]];
			    my $jB = $ap2spB[$sp2apA[$jA]];
			    
			    if ($iB != -1 && $jB != -1) {
				# print "$iA,$jA <--> $sp2apA[$iA],$sp2apA[$jA] <--> $iB,$jB\n";
				my $pB = (defined $pairprobsB[$iB][$jB])?$pairprobsB[$iB][$jB]:$min_prob;
				if ($pB>$min_prob) {
				    my $new_pA = exp( log($pA) * (1/(1+$pB)));
				    $ext_pairprobsA[$iA][$jA] = $new_pA;
				    my @orig_pairprobsA=@{ $pairprobs[$a] };
				    my $orig_pA  = $orig_pairprobsA[$iA][$jA];
				    print "Enhance ($iA,$jA):$pA by ($iA,$jB):$pB --> $new_pA/$orig_pA\n";
				}
			    }
			}
		    }
		}
	    }
	}
    }
    
    ## --------------------------------------------------
    ## write pp files with extended probabilities
    #
    
    for (my $a=0; $a<=$#names; $a++) {
	# print "write extended probs for $names[$a]\n";

	my $seq=$sequences[$a];
	my $len=length($seq);
	my @ext_pairprobs = @{ $ext_pairprobs[$a] }; # the extended probs
	

	open(PP_OUT, ">$names[$a].pp") || die "Cannot write to $names[$a].pp\n"; 
	
	print PP_OUT "$names[$a] $seq\n\n#\n";
	
	for (my $i=0; $i<$len; $i++) {
	    for (my $j=$i+1; $j<$len; $j++) {
		if (defined $ext_pairprobs[$i][$j]) {
		    my $p = $ext_pairprobs[$i][$j];
		    if ($p > $min_prob) {
			print PP_OUT "$i $j $p\n";
		    }
		}
	    }
	}
    }
    
    #
    # print "\n\n";
    #

    for (my $a=0; $a<=$#names; $a++) {
	rename "$names[$a].pp", "$names[$a]";
    }
    
    #
    ##
    ## --------------------------------------------------
}

## compute alignment
## with consistency based transformation
sub compute_alignment_from_dps_cbt($$$$$) {
    my ($seqA_dp,$seqB_dp,$locarna_params,$bmp_ref,$tmpprefix) =  @_;
   
    local *CA_IN;

    my $tmpfile="$tmpprefix.clustal";
    my $tmpfile2="$tmpprefix.bmprobs";

    write_sparse_matrix($bmp_ref,$tmpfile2);

    $locarna_params .= " --clustal=$tmpfile --read-match-probs=$tmpfile2";
    
    my $cmd="locarna $seqA_dp $seqB_dp $locarna_params >/dev/null";
    system $cmd || die "Command $cmd failed."; 
    
    open(CA_IN,"$tmpfile") || die "compute_alignment: Cannot read temporary file $tmpfile\n";
    
    my @content=<CA_IN>;
    
    close CA_IN;
    
    unlink $tmpfile;
    
    return @content;
}


### ============================================================================
### ============================================================================
### ============================================================================
### MAN-PAGE
###


__END__

=head1 NAME

MLocarna - local multiple alignment of RNA

=head1 SYNOPSIS

mlocarna [options] <file>

Options:
    --help                        Brief help message
    --man                         Full documentation
    -v, --verbose                 Verbose
    --moreverbose                 Even more verbose
    -q, --quiet                   Quiet
    -P, --tree-min-prob=<f>       Minimal prob for constructing guide tree
    --noLP                        No lonely pair
    --LP                          Allow lonely pairs
    --treefile                    Guide tree file
    --tgtdir                      Target directory
    --skip-pp                     Skip computation of pair probs if already existing
    --extlib                      Use library extension -- experimental feature
    --iterate                     Refine iteratively after progressive alignment (experimental)
    --iterations                  Refine iteratively for given number of iterations (experimental) 
    --ignore-constraints          Ignore constraints even if given
    --consistency-transformation  Apply consistency transformation (for mea, experimental)
    
    as well as most of the options of locarna (see man page of locarna)
    
=head1 DESCRIPTION

B<MLocarna> computes a multiple global or local sequence-structure alignment
of the RNA molecules that are given in input file <file> in mfasta format.
Results are written to directory <dir>. If the file tree is given, contained
tree (in NEWICK-tree format) is used as guide tree for the progressive
alignment. 

Whenever parameters are not specified explicitly, we use the locarna defaults,
except for min-prob (0.0005) and tree-min-prob (0.01).

=head1 SEE ALSO

The  full  documentation  for  LocARNA  and mLocARNA is maintained as a Texinfo manual.  If the info and
LocARNA programs are properly installed at your site, the command

    info locarna

should give you access to the complete manual.

=cut
