/**********************************************************************
 *
 * LocARNA: LOCal Alignment of RNA
 *
 * Copyright (C) Sebastian Will <will(@)informatik.uni-freiburg.de> 
 *               2005, 2006, 2007
 *
 **********************************************************************/


#include <iostream>
#include <fstream>
// #include <sstream>
#include <vector>

#include <math.h>

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"
#include "aligner.hh"

#include "rna_data.hh"

#include "match_probs.hh"
#include "ribosum.hh"

#include "anchor_constraints.hh"

using namespace std;

const std::string 
VERSION_STRING = (std::string)PACKAGE_STRING; 

// ------------------------------------------------------------
// Parameter

double min_prob; // only pairs with a probability of at least min_prob are taken into account

int match_score;
int mismatch_score;
int indel_score;
int indel_opening_score;

int struct_weight;

int tau_factor; // contribution of sequence similarity in an arc match (in percent)
               
bool no_lonely_pairs; // no lonely pairs option

bool struct_local; // allow exclusions for maximizing alignment of connected substructures 
bool sequ_local; // maximize alignment of subsequences

const bool DO_TRACE=true;


int max_diff; // maximal difference for positions of alignment edges
              // (only used for ends of arcs)
int max_diff_am; //maximal difference between two arc ends, -1 is off

// only consider arc matchs where
//   1. for global (bl-al)>max_diff || (br-ar)<=max_diff    (if max_diff>=0)
//   2. for local (ar-al)-(br-bl)<=max_diff_am              (if max_diff_am>=0)
// except when there is no additional computation of M matrices necessary,
// this occurs if arcs are left-incident with larger arcs where 1 and 2 hold


int exclusion_score; // Score contribution per exclusion
                     // set to zero for unrestricted structure locality

double prob_exp_f(int seqlen) {return 1.0/(2*seqlen);} // expected probability of a base pair (null-model)

double exp_prob;
bool opt_exp_prob;

int output_width;

// ------------------------------------------------------------
// File arguments
char *bpsfile1;
char *bpsfile2;

char *clustal_out;
bool opt_clustal_out;

char *pp_out;
bool opt_pp_out;

// ------------------------------------------------------------
//
// Options
//
#include "options.hh"


bool opt_help;
bool opt_version;
bool opt_verbose;
bool opt_local_output;
bool opt_pos_output;

bool opt_stacking;

bool opt_ribosum_file;
char *ribosum_file;
bool opt_probcons_file;
char *probcons_file;

bool opt_mea_alignment;

bool opt_write_matchprobs;
bool opt_read_matchprobs;
char *matchprobs_file;

int match_prob_method;
int temperature;

int kbest_k;

char *seq_constraints_A;
char *seq_constraints_B;

bool opt_ignore_constraints;


int mea_alpha;
int mea_beta;
int mea_gamma;


option_def my_options[] = {
    {"help",'h',&opt_help,O_NO_ARG,0,0,0,"This help"},
    {"version",'V',&opt_version,O_NO_ARG,0,0,0,"Version info"},
    {"verbose",'v',&opt_verbose,O_NO_ARG,0,0,0,"Verbose"},
    
    {0,0,0,O_SECTION,0,0,0,"Scoring parameters"},
    
    {"match",'m',0,O_ARG_INT,&match_score,"50","score","Match score"},
    {"mismatch",'M',0,O_ARG_INT,&mismatch_score,"0","score","Mismatch score"},
    {"ribosum-file",0,&opt_ribosum_file,O_ARG_STRING,&ribosum_file,0,"file","Ribosum file for base and arc match scores"},
    {"indel",'i',0,O_ARG_INT,&indel_score,"-100","score","Indel score"},
    {"indel-opening",0,0,O_ARG_INT,&indel_opening_score,"-200","score","Indel opening score"},
    {"struct-weight",'s',0,O_ARG_INT,&struct_weight,"100","score","Maximal weight of 1/2 arc match"},
    {"exp-prob",'e',&opt_exp_prob,O_ARG_DOUBLE,&exp_prob,0,"prob","Expected probability"},
    {"tau",'t',0,O_ARG_INT,&tau_factor,"0","factor","Tau factor in percent"},
    {"exclusion",'E',0,O_ARG_INT,&exclusion_score,"0","score","Exclusion weight"},
    {"stacking",0,&opt_stacking,O_NO_ARG,0,0,0,"Alignment with stacking terms (needs stack-probs. as by RNAfold -p2)"},   

    {0,0,0,O_SECTION,0,0,0,"Type of locality"},
    
    {"struct-local",0,0,O_ARG_BOOL,&struct_local,"false","bool","Structure local"},
    {"sequ-local",0,0,O_ARG_BOOL,&sequ_local,"false","bool","Sequence local"},

    {0,0,0,O_SECTION,0,0,0,"Controlling output"},
    
    {"width",'w',0,O_ARG_INT,&output_width,"120","columns","Output width"},
    {"clustal",0,&opt_clustal_out,O_ARG_STRING,&clustal_out,0,"file","Clustal output"},
    {"pp",0,&opt_pp_out,O_ARG_STRING,&pp_out,0,"file","PP output"},
    {"local-output",'L',&opt_local_output,O_NO_ARG,0,0,0,"Output only local sub-alignment"},
    {"pos-output",'P',&opt_pos_output,O_NO_ARG,0,0,0,"Output only local sub-alignment positions"},
    
    {0,0,0,O_SECTION,0,0,0,"Heuristics for speed accuracy trade off"},
    
    {"min-prob",'p',0,O_ARG_DOUBLE,&min_prob,"0.005","prob","Minimal probability"},
    {"max-diff-am",'D',0,O_ARG_INT,&max_diff_am,"-1","diff","Maximal difference for sizes of matched arcs"},
       
    {0,0,0,O_SECTION,0,0,0,"Special sauce options"},
    {"kbest",'k',0,O_ARG_INT,&kbest_k,"-1","k","Find k-best alignments"},
    
    {0,0,0,O_SECTION,0,0,0,"Options for controlling MEA score (experimental, under construction)"},
    
    {"mea-alignment",0,&opt_mea_alignment,O_NO_ARG,0,0,0,"Do MEA alignment"},
    {"probcons-file",0,&opt_probcons_file,O_ARG_STRING,&probcons_file,0,"file","Probcons parameter file"},
    
    {"match-prob-method",0,0,O_ARG_INT,&match_prob_method,"0","int","Method for computation of match probs"},
    {"temperature",0,0,O_ARG_INT,&temperature,"800","int","Temperature for PF-computation"},
    {"mea-alpha",0,0,O_ARG_INT,&mea_alpha,"100","weight","Weight alpha for MEA"},
    {"mea-beta",0,0,O_ARG_INT,&mea_beta,"100","weight","Weight beta for MEA"},
    {"mea-gamma",0,0,O_ARG_INT,&mea_gamma,"100","weight","Weight gamma for MEA"},
    
    {"write-match-probs",0,&opt_write_matchprobs,O_ARG_STRING,&matchprobs_file,0,"file","Write match probs to file (don't align!)"},
    {"read-match-probs",0,&opt_read_matchprobs,O_ARG_STRING,&matchprobs_file,0,"file","Read match probabilities from file"},

    {0,0,0,O_SECTION,0,0,0,"Constraints"},
    
    {"noLP",0,&no_lonely_pairs,O_NO_ARG,0,0,0,"No lonely pairs"},
    {"anchorA",0,0,O_ARG_STRING,&seq_constraints_A,"","string","Anchor constraints sequence A"},
    {"anchorB",0,0,O_ARG_STRING,&seq_constraints_B,"","string","Anchor constraints sequence B"},
    {"ignore-constraints",0,&opt_ignore_constraints,O_NO_ARG,0,0,0,"Ignore constraints in pp-file"},
    
    {0,0,0,O_SECTION_HIDE,0,0,0,"Experimental options (partially incompleted, immature, no guarantees)"},
    
    {"max-diff-match",'d',0,O_ARG_INT,&max_diff,"-1","diff","Maximal difference for alignment edges"},
    
    {0,0,0,O_SECTION,0,0,0,"RNA sequences and pair probabilities"},
    
    {0,0,0,O_ARG_STRING,&bpsfile1,0,"bps-file 1","Basepairs input file 1"},
    {0,0,0,O_ARG_STRING,&bpsfile2,0,"bps-file 2","Basepairs input file 2"},
    {0,0,0,0,0,0,0,0}
};


// ------------------------------------------------------------


/* ------------------------------------------------------------
   Experimental feature: adapt score to sequence identity
   ------------------------------------------------------------ */

/*
prob_equals returns "probability" that two alignment columns are equal
this function needs to be re-implemented

double seq_identity(Sequence a, Sequence b) {
    const size_t n=a.length();
    const size_t m=b.length();

    vector<vector<double> > M;

    M.resize(n+1);
    for (uint i=0; i<=n; i++) M[i].resize(m+1);
        
    for (uint i=0; i<=n; i++) M[i][0]=0;
    for (uint j=0; j<=m; j++) M[0][j]=0;

    for (uint i=1; i<=n; i++)
	for (uint j=1; j<=m; j++)
	    M[i][j] = max(M[i-1][j-1] + a[i].prob_equals(b[j]),
			  max(M[i-1][j],
			      M[i][j-1]));
			      
    return 2.0*M[n][m]/(double)(n+m);
}
*/


/* ------------------------------------------------------------ */


// ------------------------------------------------------------
// MAIN

int
main(int argc, char **argv) {
    
    bool process_success=process_options(argc,argv,my_options);

    if (opt_help) {
	cout << VERSION_STRING<<endl;
	
	cout << "Copyright Sebastian Will, 2004-2008"<<endl<<endl;
	
	cout << "A tool for pairwise Local (and global) Alignment of RNA."<<endl<<endl;
	
	print_help(argv[0],my_options);
	
	cout << "Report bugs to <will (at) informatik.uni-freiburg.de>."<<endl<<endl;
	exit(0);
    }
    
    if (opt_version || opt_verbose) {
	cout << VERSION_STRING<<endl;
	if (opt_version) exit(0); else cout <<endl;
    }

    if (!process_success) {
	puts(O_error_msg);
	printf("USAGE: ");
	print_usage(argv[0],my_options);
	printf("\n");
	exit(-1);
    }
        
    if (opt_verbose)
	print_options(my_options);
    


    BasePairsParams bp_params(min_prob);

    RnaData rnadataA(bpsfile1,opt_stacking);
    RnaData rnadataB(bpsfile2,opt_stacking);

    Sequence seqA=rnadataA.get_sequence();
    Sequence seqB=rnadataB.get_sequence();
    
    BasePairs bpsA(&rnadataA,&bp_params);
    BasePairs bpsB(&rnadataB,&bp_params);
    
    if (opt_verbose) {
	cout << "Sequence A: "<<endl;
	seqA.write(cout);
	cout<<" (Length:"<< seqA.length()<<", Basepairs:"<<bpsA.num_bps() << ")" <<endl;
		
	cout << "Sequence B: "<<endl;
	seqB.write(cout);
	cout<<" (Length:"<< seqB.length()<<", Basepairs:"<<bpsB.num_bps() << ")" <<endl;

	// cout << "Base Identity: "<<(seq_identity(seqA,seqB)*100)<<endl; 
    }

    // ----------------------------------------  

    Ribosum *ribosum=NULL;
    
    if (opt_ribosum_file) {
	ribosum = new RibosumFreq(ribosum_file);
    }
    
    MatchProbs *match_probs=0L;

    if (opt_read_matchprobs && !opt_mea_alignment) {
	std::cerr << "Warning: opt_read_matchprobs ignored for non-mea alignment.\n"; 
    }
    
    if (opt_write_matchprobs || opt_mea_alignment) {
	match_probs = new MatchProbs;
	
	if (!opt_ribosum_file) {
	    std::cerr << "Ribosum file required for mea_alignment and computing matchprobs."<<std::endl;
	    print_usage(argv[0],my_options);
	    std::cerr << std::endl;
	    exit(-1);
	}
	
	if (opt_read_matchprobs) {
	    match_probs->read_sparse(matchprobs_file,seqA.length(),seqB.length());
	} else {
	    if (match_prob_method==1) {
		if (!opt_probcons_file) {
		    std::cerr << "Probcons parameter file required for pairHMM-style computation"
			      <<" of basematch probabilities."<<std::endl;
		    print_usage(argv[0],my_options);
		    std::cerr << std::endl;
		    exit(-1);
		}
		if (opt_verbose) {
		    std::cout << "Compute match probabilities using pairHMM."<<std::endl; 
		}
		
		match_probs->pairHMM_probs(seqA,seqB,probcons_file);
	    } else {
		bool sl=sequ_local;
		if (match_prob_method==2) sl=true;
		if (match_prob_method==3) sl=false;
		
		if (opt_verbose) {
		    std::cout << "Compute match probabilities using PF alignment."<<std::endl; 
		}

		match_probs->pf_probs(seqA,seqB,
				      ribosum->get_basematch_scores(),
				      ribosum->alphabet(),
				      indel_opening_score/100.0,indel_score/100.0,temperature/100.0,sl);
	    }
	}

	if (opt_write_matchprobs) {
	    if (opt_verbose) {
		std::cout << "Write match probabilities to file "<<matchprobs_file<<"."<<std::endl; 
	    }
	    
	    match_probs->write_sparse(matchprobs_file,1.0/Scoring::score_resolution);
	    exit(0);
	}
    }
    
    
    ScoringParams scoring_params(match_score,
				 mismatch_score,
				 opt_mea_alignment?0:indel_score, // in true mea alignment gap are only scored for computing base match probs
				 opt_mea_alignment?0:indel_opening_score,
				 ribosum,
				 struct_weight,
				 tau_factor,
				 exclusion_score,
				 opt_exp_prob?exp_prob:-1,
				 opt_stacking,
				 opt_mea_alignment,
				 match_probs,
				 mea_alpha,
				 mea_beta,
				 mea_gamma
	);
    
    Scoring scoring(&bpsA,&bpsB,&scoring_params);
    
    std::string seqCA = seq_constraints_A;
    std::string seqCB = seq_constraints_B;
    
    if (!opt_ignore_constraints) {
	if ( seqCA=="" ) seqCA = rnadataA.get_seq_constraints();
	if ( seqCB=="" ) seqCB = rnadataB.get_seq_constraints();
    }
    
    AnchorConstraints seq_constraints(seqA.length(),seqCA,
				   seqB.length(),seqCB);
    
    if (opt_verbose) {
	if (! seq_constraints.empty()) {
	    std::cout << "Found sequence constraints."<<std::endl;
	}
    }
   
    AlignerParams aligner_params(
	no_lonely_pairs,
	struct_local, sequ_local,
	max_diff, max_diff_am,
	opt_stacking,
	seq_constraints
	);
    
    Aligner aligner(seqA,bpsA,seqB,bpsB,&aligner_params,&scoring);

    // if option --kbest is given, use kbest algo (by interval splitting)
    if (kbest_k>0) {
	aligner.kbest(kbest_k,
		      output_width,
		      opt_verbose,
		      opt_local_output,
		      opt_pos_output
		      );
	exit(0);
    }


    // otherwise find the best alignment
	
    infty_score_t score = aligner.align();
    
    std::cout << "Score: "<<score<<std::endl;
    
    if (DO_TRACE) {

	 aligner.trace();
	 aligner.get_alignment().write(std::cout, output_width, score,
				       opt_local_output, opt_pos_output);
	 std::cout<<endl;

	 if (opt_clustal_out) {
	     ofstream out(clustal_out);
	     if (out) {
		 aligner.get_alignment().write_clustal(out, output_width,
						       score,
						       opt_local_output,
						       opt_pos_output);
	    } else {
		cerr << "Cannot write to "<<clustal_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
	if (opt_pp_out) {
	    ofstream out(pp_out);
	    if (out) {
		aligner.get_alignment().
		    write_pp(out,bpsA,bpsB,scoring,seq_constraints,output_width);
	    } else {
		cerr << "Cannot write to "<<pp_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
    }
    exit(0);
}
