#ifndef RNA_DATA_HH
#define RNA_DATA_HH

#include <algorithm>
#include <string>
#include <iostream>

#include "sequence.hh"

#include "sparse_matrix.hh"

//! Represents the input data per RNA,
//! i.e. the set of base pairs and the RNA sequence
//!
//! Reads, maintains and provides access to
//! the set of basepairs of one RNA
//! together with their pair probabilities
//! Features: read from pp or dp_ps
//! also maintains stacking probabilities
//! 
//! possible extensions: 
//! offer predicition using Vienna RNA lib
//!
//! supports the definition of sequence constraints in pp files

class RnaData {
public:
    typedef Sequence::size_type size_type;
    typedef SparseMatrix<double> arc_prob_matrix_t;
private:
    Sequence sequence;
    bool stacking;

    arc_prob_matrix_t arc_probs_; //!< array for all arc probabilities
    //!< the array is used
    //!< when reading in the probabilities
    //!< and for merging probs during pp-output
    arc_prob_matrix_t arc_2_probs_; //!< array for all probabilities
                                    //!< that a pair (i,j) and its immediately inner pair (i+1,j-1) 
                                    //!< are formed simultatneously;
                                    //!< analogous to arc_probs_
    
    std::string seq_constraints_;


    class ToUpper { 
    public:
	char operator() (char c) const  { return std::toupper(c); }
    };
    static void transform_toupper(std::string &s) {
	std::transform(s.begin(),s.end(),s.begin(),ToUpper());
    }


public:
    //! construct by reading basepairs from file (pp or dp_ps)
    RnaData(const std::string &file, bool stacking=false);
    
    // construct from sequence, predicting the basepairs
    RnaData(const Sequence &sequence_)
	: sequence(sequence_),
	  arc_probs_(0),
	  arc_2_probs_(0)
	{
	    std::cerr << "construct RnaData from sequence, currently not implemented."
		      << std::endl;
	    exit(-1);
	}
    
    //! get the sequence
    const Sequence &get_sequence() const {
	return sequence;
    }
    
    //! get the sequence
    const std::string &get_seq_constraints() const {
	return seq_constraints_;
    }
    
    
private:
    // ------------------------------------------------------------
    // reading methods
    
    //! read basepairs and sequence from a pp-format file
    //! pp is a proprietary format of LocARNA
    //! which starts with the sequence and then simply
    //! lists the arcs (i,j) with their probabilities p
    void readPP(const std::string &filename);
    
    //! read baepairs and sequence from a dp-format ps file
    //! dp is written by RNAfold -p
    void readPS(const std::string &filename);
    
    //! read baepairs and sequence from a file
    //! (try to autodetect file format)
    //!@param filename the input file
    //!@param seq output parameter for the sequence
    //!@param stacking reads also stacking probabilities
    void read(const std::string &filename);
    
    
    // ------------------------------------------------------------
    // set methods
    
    //! sets probability of basepair (i,j)
    void set_arc_prob(int i, int j, double p) {
	assert(i<j); 
	arc_probs_.set(i,j,p);
    }

    //! sets stacking probability of basepairs (i,j) and (i+1,j-1)
    void set_arc_2_prob(int i, int j, double p) {
	assert(i<j);
	arc_2_probs_.set(i,j,p);
    }
    
public:
    // ------------------------------------------------------------
    // get methods
    
    //! returns probability of basepair (i,j)
    double get_arc_prob(size_type i, size_type j) const {
	assert(i<j);
	return arc_probs_(i,j);
    }


    //! returns probability of basepairs (i,j) and (i+1,j-1) occuring simultaneously
    double get_arc_2_prob(size_type i, size_type j) const {
	assert(i<j); 
	return arc_2_probs_(i,j);
    }

    double get_arc_stack_prob(size_type i, size_type j) const {
	assert(i<j); 
	return arc_2_probs_(i,j)/get_arc_prob(i+1,j-1);
    }
    
    // ------------------------------------------------------------
    // compute unpaired probability
    
    //! returns probability that a position i is unpaired
    //! ATTENTION: expensive implementation,
    //! could be replaced by table lookup
    double prob_unpaired(size_type i) const {
	double prob_paired=0.0;
	
	for (size_type j=1; j<i; j++) {
	    prob_paired += arc_probs_(j,i); 
	}
	for (size_type j=i+1; j<=sequence.length(); j++) {
	    prob_paired += arc_probs_(i,j); 
	}
	
	return 1.0-prob_paired;
    }
    
    // ------------------------------------------------------------
    // misc
    
    std::string seqname_from_filename(const std::string &s) const;
};


#endif // RNA_DATA_HH
