#ifndef SCORING_HH
#define SCORING_HH

#include <math.h>

#include "matrices.hh"
#include "basepairs.hh"
#include "match_probs.hh"
#include "rna_data.hh"
#include "ribosum.hh"


#include "infty_arith_int.hh"

typedef long int score_t;
typedef InftyArithInt infty_score_t;


class Ribosum;
class Scoring;

//! Contains all parameters
//! for doing the scoring of alignments
//! in class Scoring.
//! The class encapsulates the configuration
//! of the score.
//!
class ScoringParams {
    friend class Scoring;

    //! constant bonus for a base match.
    //! together with basemismatch yields the simplest form
    //! of base match/mismatch scoring.
    //! Can be replaced by RIBOSUM scores (planned).
    score_t basematch;
    
    //! constant cost of a base mismatch
    score_t basemismatch;
   
    //! cost per indel (for linear or affine gap cost).
    score_t indel;

    //! cost per gap (for affine gap-cost). Use affine gap cost if non-zero.
    score_t indel_opening;
        
    //! the ribosum matrix, if non-null it is used
    //! for base match/mismatch instead of constant values
    //! and as contribution for arc-matchs (tau_factor)
    Ribosum *ribosum;
    
    //! Factor for structure contribution in the classic score.
    //! Maximal contribution of 1/2 arc match
    score_t struct_weight;
    
    //! Factor for the contribution of sequence score or
    //! ribosum score to arc matchs
    score_t tau_factor;
    
    //! cost of one exclusion.
    score_t exclusion;

    double exp_prob;

    //! turn on/off stacking terms
    bool stacking;

    //! turn on/off mea scoring
    bool mea_scoring;

    //! probabilities of base matches in sequence alignment
    MatchProbs *match_probs;

    //! weight for mea contribution "unstructured"
    score_t alpha_factor;

    //! weight for mea contribution "structure" 
    score_t beta_factor;

    //! weight for mea contribution "consensus"
    score_t gamma_factor;

    /*
      The mea score is composed in the following way:
      
      score_resolution *
      (
       sum_basematchs (i,j)
           P(i~j)
           + alpha_factor/100 * (P(i unstructured) + P(j unstructured))
       + 
       sum_arcmatchs (a,b)
           beta_factor/100 * (P(a)+P(b))
	   gamma_factor/100 * (ribosum_arcmatch(a,b))
      )
    */
    
public:
    ScoringParams(score_t basematch_,
		  score_t basemismatch_,
		  score_t indel_,
		  score_t indel_opening_,
		  Ribosum *ribosum_,
		  score_t struct_weight_,
		  score_t tau_factor_,
		  score_t exclusion_,
		  double exp_prob_,
		  bool stacking_,
		  bool mea_scoring_,
		  MatchProbs *match_probs_,
		  score_t alpha_factor_,
		  score_t beta_factor_,
		  score_t gamma_factor_
		  )
	: basematch(basematch_),
	  basemismatch(basemismatch_),
	  indel(indel_),
	  indel_opening(indel_opening_),
	  ribosum(ribosum_),
	  struct_weight(struct_weight_),
	  tau_factor(tau_factor_),
	  exclusion(exclusion_),
	  exp_prob(exp_prob_),
	  stacking(stacking_),
	  mea_scoring(mea_scoring_),
	  match_probs(match_probs_),
	  alpha_factor(alpha_factor_),
	  beta_factor(beta_factor_),
	  gamma_factor(gamma_factor_)
	{
	}
};

//! Maintains and provides the scores for the alignment of two specific
//! Rnas.
//! Highly configurable via class ScoringParams, supports
//! the classic LocARNA-score as well as the LocARNA MEA-Score.
//! for "Classic Score" it supports
//!  * match/mismatch cost or RIBOSUM (planned) for bases
//!  * linear or affine gap cost
//!  * arc match scores build of arc probabilities
//!    and optionally base match contribution or RIBOSUM (planned)
//! for the "MEA Score" it supports
//!  * base match/mismatch scores derived from base match probabilities
//!  * RIBOSUM for base and arc match contribution
//!  * arc match score contributions from arc probabilities
//!  * weighting of single score contributions
class Scoring {
    
public:
    //! resolution of the mea score
    //! (since the mea score is composed from
    //! probabilities and scores are integral
    //! there must be some factor to scale the score)
    static const score_t score_resolution=10000;

private:
    const ScoringParams *params;
    const BasePairs *bpsA;
    const BasePairs *bpsB;
    
    const Alphabet<char> &alphabet;
    
public:
    typedef Sequence::size_type size_type;

    Scoring(const BasePairs *bpsA_,
	    const BasePairs *bpsB_,
	    const ScoringParams *params_):
	params(params_),
	bpsA(bpsA_),
	bpsB(bpsB_),
	alphabet(Sequence::alphabet())
	{
	    // std::cout << "sizeof score_t = "<<sizeof(score_t)<<std::endl;
	    precompute_sigma();
	    precompute_gapcost();
	    precompute_weights();
	}
    
private:
    Matrix<score_t> sigma_tab;
    
    std::vector<score_t> gapcost_tabA; //! table for gapcost in A
    std::vector<score_t> gapcost_tabB; //! table for gapcost in B
    
    std::vector<score_t> weightsA;
    std::vector<score_t> weightsB;

    std::vector<score_t> stack_weightsA;
    std::vector<score_t> stack_weightsB;
    
    //! compute base similarity
    score_t
    sigma_(int i, int j) const;

    //! precompute all base similarities
    //! and put them in the sigma table
    void
    precompute_sigma();  

    //! precompute the tables for gapcost
    void
    precompute_gapcost();
    
    //! compute weights/stacked weights for all arcs in A and B
    void
    precompute_weights();

    //! helper for precompute_weights (does job for one rna)
    void
    precompute_weights(const BasePairs &bps,
		       std::vector<score_t> &weights,
		       std::vector<score_t> &stack_weights);
    
    //! returns weight that corresponds to probability p
    //! in non-mea score
    score_t
    probToWeight(double p, size_t len) const;
    
    double
    prob_exp_f(int seqlen) const {return 1.0/(2.0*seqlen);}//!<expected probability of a base pair (null-model)
    

    //! returns probability of matching the concrete bases in an arcmatch,
    //! probability is based on ribosum data
    double
    ribosum_arcmatch_prob(const Arc &arcA, const Arc &arcB) const;    

    score_t
    ribosum_arcmatch_score(const Arc &arcA, const Arc &arcB) const;

public:
    // ------------------------------------------------------------
    // SCORE CONTRIBUTIONS
    
    //! score an match of bases (without structure)
    score_t basematch(size_t i, size_t j) const {
	return sigma_tab(i,j);
    }
    
    //! score arc match
    score_t arcmatch(const Arc &arcA, const Arc &arcB, bool stacked=false) const {
	double ribosum_contribution=0;
	score_t sequence_contribution=0;


	// for the mea score, we need the probability of the arc match
	// from the sequence information
	if ( params->mea_scoring ) {
	    ribosum_contribution = ribosum_arcmatch_prob(arcA,arcB);
	}

	// if tau is non-zero
	// we add a sequence contribution
	// for mea or if we don't have a ribosum matrix, we use sigma
	// otherwise we use a ribosum-like score 
	if ( params->tau_factor != 0 ) {
	    if (!params->mea_scoring && params->ribosum) {
		sequence_contribution = ribosum_arcmatch_score(arcA,arcB);
	    } else {
		sequence_contribution =
		    sigma_tab(arcA.left(),arcB.left())
		    + sigma_tab(arcA.right(),arcB.right());
	    }
	}
	
	return
	    static_cast<score_t>
	    (
		(
		    // base match contribution
		    // (also for arc-match we add terms for the base match on both ends, weighted by tau_factor)
		    (params->tau_factor/100.0) * sequence_contribution
		    +
		    // structure contribution, weighted by beta
		    (params->beta_factor/100.0) 
		    * 
		    ( 
			stacked
			?
			(stack_weightsA[arcA.idx()] + stack_weightsB[arcB.idx()])
			:
			(weightsA[arcA.idx()] + weightsB[arcB.idx()])
			)
		    +
		    // consensus contribution
		    score_resolution *
		    (params->gamma_factor/100.0) *
		    (
			params->match_probs==NULL
			?
			1.0
			:
			(
			    params->match_probs->prob(arcA.left(),arcB.left())
			    * params->match_probs->prob(arcA.right(),arcB.right())
			    )
			)
		    *
		    ribosum_contribution
		    )
		);
    }
    
    //! stacked arc match
    score_t arcmatch_stacked(const Arc &arcA, const Arc &arcB) const {
	return arcmatch(arcA, arcB, true);
    }

    //! score a gap (only linear gap cost)
    //! if (seq==0): score of deletion of posA after posB
    //! if (seq==1): score of insertion of posB after posA
    //!
    score_t gapA(size_type posA, size_type posB) const {
	return gapcost_tabA[posA];
    }
    score_t gapB(size_type posA, size_type posB) const {
	return gapcost_tabB[posB];
    }
    
    //! cost of an exclusion
    score_t exclusion() const {
	return params->exclusion;
    }   

    //! cost to begin a new indel
    score_t indel_opening() const {
	return params->indel_opening;
    }
    
    //
    // ------------------------------------------------------------
    
    //! expected basepair probability,
    //! influences computation of weights for basepairs
    double prob_exp(size_t len) const; //!< return expected probability

    bool stacking() const {return params->stacking;}

};

#endif // SCORING_HH
